import argparse
from types import SimpleNamespace

from fdfat.utils.logger import LOGGER
from fdfat.cfg import get_cfg
# from fdfat.main import TrainEngine, ValEngine, TestEngine, 
from fdfat.engine import trainer, validator, predictor, exporter

def parse_bool(v):
    return v.lower() in ("yes", "true", "t", "1")

def entrypoint():
    default_cfg = get_cfg()

    parser = argparse.ArgumentParser(description='train loop.')
    for k, v in default_cfg:
        t = type(v)
        t = parse_bool if t == bool else (t if v is not None else str)
        parser.add_argument(f"--{k}", type=t, default=v)
    args = parser.parse_args()

    print(args)

    args = SimpleNamespace(**vars(args))
    
    if args.task == "train":
        engine = trainer.TrainEngine(args)
        engine.prepare()
        engine.do_train()
    elif args.task == "val":
        engine = validator.ValEngine(args)
        engine.prepare()
        engine.do_validate()
    elif args.task == "predict":
        engine = predictor.PredictEngine(args)
        if args.input is None:
            raise ValueError("Input is empty")
        lmk, rendered = engine.predict(args.input, render=True)
        rendered.save("predict.jpg")
    elif args.task == "export":
        engine = exporter.ExportEngine(args)
        engine.export()



import os
import logging.config
import pkg_resources as pkg

from fdfat import LOGGING_NAME, VERBOSE, MACOS, LINUX, WINDOWS

def check_version(current: str = '0.0.0',
                  minimum: str = '0.0.0',
                  name: str = 'version ',
                  pinned: bool = False,
                  hard: bool = False,
                  verbose: bool = False) -> bool:
    """
    Check current version against the required minimum version.

    Args:
        current (str): Current version.
        minimum (str): Required minimum version.
        name (str): Name to be used in warning message.
        pinned (bool): If True, versions must match exactly. If False, minimum version must be satisfied.
        hard (bool): If True, raise an AssertionError if the minimum version is not met.
        verbose (bool): If True, print warning message if minimum version is not met.

    Returns:
        (bool): True if minimum version is met, False otherwise.
    """
    current, minimum = (pkg.parse_version(x) for x in (current, minimum))
    result = (current == minimum) if pinned else (current >= minimum)  # bool
    warning_message = f'WARNING ⚠️ {name}{minimum} is required, but {name}{current} is currently installed'
    if hard:
        assert result, emojis(warning_message)  # assert min requirements met
    if verbose and not result:
        LOGGER.warning(warning_message)
    return result

def set_logging(name=LOGGING_NAME, verbose=True):
    """Sets up logging for the given name."""
    rank = int(os.getenv('RANK', -1))  # rank in world for Multi-GPU trainings
    level = logging.INFO if verbose and rank in {-1, 0} else logging.ERROR
    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            name: {
                'format': '%(message)s'}},
        'handlers': {
            name: {
                'class': 'logging.StreamHandler',
                'formatter': name,
                'level': level}},
        'loggers': {
            name: {
                'level': level,
                'handlers': [name],
                'propagate': False}}})

def emojis(string=''):
    """Return platform-dependent emoji-safe version of string."""
    return string.encode().decode('ascii', 'ignore') if WINDOWS else string

class EmojiFilter(logging.Filter):
    """
    A custom logging filter class for removing emojis in log messages.

    This filter is particularly useful for ensuring compatibility with Windows terminals
    that may not support the display of emojis in log messages.
    """

    def filter(self, record):
        """Filter logs by emoji unicode characters on windows."""
        record.msg = emojis(record.msg)
        return super().filter(record)

# Set logger
set_logging(LOGGING_NAME, verbose=VERBOSE)  # run before defining LOGGER
LOGGER = logging.getLogger(LOGGING_NAME)  # define globally (used in train.py, val.py, detect.py, etc.)
if WINDOWS:  # emoji-safe logging
    LOGGER.addFilter(EmojiFilter())

def colorstr(*input):
    """Colors a string https://en.wikipedia.org/wiki/ANSI_escape_code, i.e.  colorstr('blue', 'hello world')."""
    *args, string = input if len(input) > 1 else ('blue', 'bold', input[0])  # color arguments, string
    colors = {
        'black': '\033[30m',  # basic colors
        'red': '\033[31m',
        'green': '\033[32m',
        'yellow': '\033[33m',
        'blue': '\033[34m',
        'magenta': '\033[35m',
        'cyan': '\033[36m',
        'white': '\033[37m',
        'bright_black': '\033[90m',  # bright colors
        'bright_red': '\033[91m',
        'bright_green': '\033[92m',
        'bright_yellow': '\033[93m',
        'bright_blue': '\033[94m',
        'bright_magenta': '\033[95m',
        'bright_cyan': '\033[96m',
        'bright_white': '\033[97m',
        'end': '\033[0m',  # misc
        'bold': '\033[1m',
        'underline': '\033[4m'}
    return ''.join(colors[x] for x in args) + f'{string}' + colors['end']


import re
import yaml
from pathlib import Path
from typing import Union
from types import SimpleNamespace

from fdfat.utils.logger import LOGGER, colorstr

FILE = Path(__file__).resolve()
ROOT = FILE.parents[0]
DEFAULT_CFG_PATH = ROOT / 'default.yaml'
DEFAULT_CFG_DATA_PATH = ROOT / 'default_data.yaml'

def yaml_save(file='data.yaml', data={}):
    """
    Save YAML data to a file.

    Args:
        file (str, optional): File name. Default is 'data.yaml'.
        data (dict): Data to save in YAML format.

    Returns:
        None: Data is saved to the specified file.
    """
    file = Path(file)
    if not file.parent.exists():
        # Create parent directories if they don't exist
        file.parent.mkdir(parents=True, exist_ok=True)

    # Convert Path objects to strings
    for k, v in data.items():
        if isinstance(v, Path):
            data[k] = str(v)

    # Dump data to file in YAML format
    with open(file, 'w') as f:
        yaml.safe_dump(data, f, sort_keys=False, allow_unicode=True)


def yaml_load(file='data.yaml', append_filename=False):
    """
    Load YAML data from a file.

    Args:
        file (str, optional): File name. Default is 'data.yaml'.
        append_filename (bool): Add the YAML filename to the YAML dictionary. Default is False.

    Returns:
        dict: YAML data and file name.
    """
    with open(file, errors='ignore', encoding='utf-8') as f:
        s = f.read()  # string

        # Remove special characters
        if not s.isprintable():
            s = re.sub(r'[^\x09\x0A\x0D\x20-\x7E\x85\xA0-\uD7FF\uE000-\uFFFD\U00010000-\U0010ffff]+', '', s)

        # Add YAML filename to dict and return
        return {**yaml.safe_load(s), 'yaml_file': str(file)} if append_filename else yaml.safe_load(s)


def yaml_print(yaml_file: Union[str, Path, dict]) -> None:
    """
    Pretty prints a yaml file or a yaml-formatted dictionary.

    Args:
        yaml_file: The file path of the yaml file or a yaml-formatted dictionary.

    Returns:
        None
    """
    yaml_dict = yaml_load(yaml_file) if isinstance(yaml_file, (str, Path)) else yaml_file
    dump = yaml.dump(yaml_dict, sort_keys=False, allow_unicode=True)
    LOGGER.info(f"Printing '{colorstr('bold', 'black', yaml_file)}'\n\n{dump}")

class IterableSimpleNamespace(SimpleNamespace):
    """
    Ultralytics IterableSimpleNamespace is an extension class of SimpleNamespace that adds iterable functionality and
    enables usage with dict() and for loops.
    """

    def __iter__(self):
        """Return an iterator of key-value pairs from the namespace's attributes."""
        return iter(vars(self).items())

    def __str__(self):
        """Return a human-readable string representation of the object."""
        return '\n'.join(f'{k}={v}' for k, v in vars(self).items())

    def __getattr__(self, attr):
        """Custom attribute access error message with helpful information."""
        name = self.__class__.__name__
        raise AttributeError(f"""
            '{name}' object has no attribute '{attr}'. This may be caused by a modified or out of date 'default.yaml' file.
            """)

    def get(self, key, default=None):
        """Return the value of the specified key if it exists; otherwise, return the default value."""
        return getattr(self, key, default)

# Default configuration
DEFAULT_CFG_DICT = yaml_load(DEFAULT_CFG_PATH)
for k, v in DEFAULT_CFG_DICT.items():
    if isinstance(v, str) and v.lower() == 'none':
        DEFAULT_CFG_DICT[k] = None
DEFAULT_CFG_KEYS = DEFAULT_CFG_DICT.keys()
DEFAULT_CFG = IterableSimpleNamespace(**DEFAULT_CFG_DICT)

DEFAULT_CFG_DATA_DICT = yaml_load(DEFAULT_CFG_DATA_PATH)
for k, v in DEFAULT_CFG_DATA_DICT.items():
    if isinstance(v, str) and v.lower() == 'none':
        DEFAULT_CFG_DATA_DICT[k] = None
DEFAULT_CFG_DATA_KEYS = DEFAULT_CFG_DATA_DICT.keys()
DEFAULT_CFG_DATA = IterableSimpleNamespace(**DEFAULT_CFG_DATA_DICT)


import os
import random
import platform
import numpy as np
from copy import deepcopy

import torch
import torch.nn as nn
import torchvision

from fvcore.nn import FlopCountAnalysis

from fdfat.utils.logger import LOGGER, check_version

TORCHVISION_0_10 = check_version(torchvision.__version__, '0.10.0')
TORCH_1_9 = check_version(torch.__version__, '1.9.0')
TORCH_1_11 = check_version(torch.__version__, '1.11.0')
TORCH_1_12 = check_version(torch.__version__, '1.12.0')
TORCH_2_0 = check_version(torch.__version__, minimum='2.0')

def is_parallel(model):
    """Returns True if model is of type DP or DDP."""
    return isinstance(model, (nn.parallel.DataParallel, nn.parallel.DistributedDataParallel))

def de_parallel(model):
    """De-parallelize a model: returns single-GPU model if model is of type DP or DDP."""
    return model.module if is_parallel(model) else model

def model_info(model, detailed=False, verbose=True, imgsz=640, device="cpu"):
    """Model information. imgsz may be int or list, i.e. imgsz=640 or imgsz=[640, 320]."""
    if not verbose:
        return
    n_p = get_num_params(model)
    n_g = get_num_gradients(model)  # number gradients
    if detailed:
        LOGGER.info(
            f"{'layer':>5} {'name':>40} {'gradient':>9} {'parameters':>12} {'shape':>20} {'mu':>10} {'sigma':>10}")
        for i, (name, p) in enumerate(model.named_parameters()):
            name = name.replace('module_list.', '')
            LOGGER.info('%5g %40s %9s %12g %20s %10.3g %10.3g %10s' %
                        (i, name, p.requires_grad, p.numel(), list(p.shape), p.mean(), p.std(), p.dtype))

    flops = get_flops(model, imgsz, device=device)
    fused = ' (fused)' if model.is_fused() else ''
    fs = f', {flops:.1f} GFLOPs' if flops else ''
    LOGGER.info(f'Summary{fused}: {len(list(model.modules()))} layers, {n_p} parameters, {n_g} gradients{fs}')
    return n_p, flops


def get_num_params(model):
    """Return the total number of parameters in a YOLO model."""
    return sum(x.numel() for x in model.parameters())


def get_num_gradients(model):
    """Return the total number of parameters with gradients in a YOLO model."""
    return sum(x.numel() for x in model.parameters() if x.requires_grad)


def get_flops(model, imgsz=640, device="cpu"):
    im = torch.empty((1, 3, imgsz, imgsz)).to(device)  # input image in BCHW format
    flops = FlopCountAnalysis(model, im)
    return flops.total()*10e-9
    # """Return a YOLO model's FLOPs."""
    # try:
    #     model = de_parallel(model)
    #     p = next(model.parameters())
    #     stride = max(int(model.stride.max()), 32) if hasattr(model, 'stride') else 32  # max stride
    #     im = torch.empty((1, p.shape[1], imgsz, imgsz), device=p.device)  # input image in BCHW format
    #     print(im.shape)
    #     flops = thop.profile(deepcopy(model), inputs=[im], verbose=False)[0] / 1E9 * 2 if thop else 0  # stride GFLOPs
    #     imgsz = imgsz if isinstance(imgsz, list) else [imgsz, imgsz]  # expand if int/float
    #     flops = flops * imgsz[0] / stride * imgsz[1] / stride  # 640x640 GFLOPs
    #     return flops
    # except Exception as e:
    #     import traceback
    #     traceback.print_exc()
    #     print(e)
    #     return 0
    
def init_seeds(seed=0, deterministic=False):
    """Initialize random number generator (RNG) seeds https://pytorch.org/docs/stable/notes/randomness.html."""
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.mps.manual_seed(seed)  # for Multi-GPU, exception safe
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)  # for Multi-GPU, exception safe
    # torch.backends.cudnn.benchmark = True  # AutoBatch problem https://github.com/ultralytics/yolov5/issues/9287
    if deterministic:  # https://github.com/ultralytics/yolov5/pull/8213
        if TORCH_2_0:
            torch.use_deterministic_algorithms(True)
            torch.backends.cudnn.deterministic = True
            os.environ['CUBLAS_WORKSPACE_CONFIG'] = ':4096:8'
            os.environ['PYTHONHASHSEED'] = str(seed)
        else:
            LOGGER.warning('WARNING ⚠️ Upgrade to torch>=2.0.0 for deterministic training.')

def select_device(device='', batch=0, newline=False, verbose=True):
    """Selects PyTorch Device. Options are device = None or 'cpu' or 0 or '0' or '0,1,2,3'."""
    s = f'Realtime landmark 🚀 Python-{platform.python_version()} torch-{torch.__version__} '
    device = str(device).lower()
    for remove in 'cuda:', 'none', '(', ')', '[', ']', "'", ' ':
        device = device.replace(remove, '')  # to string, 'cuda:0' -> '0' and '(0, 1)' -> '0,1'
    cpu = device == 'cpu'
    mps = device == 'mps'  # Apple Metal Performance Shaders (MPS)
    if cpu or mps:
        os.environ['CUDA_VISIBLE_DEVICES'] = '-1'  # force torch.cuda.is_available() = False
    elif device:  # non-cpu device requested
        visible = os.environ.get('CUDA_VISIBLE_DEVICES', None)
        os.environ['CUDA_VISIBLE_DEVICES'] = device  # set environment variable - must be before assert is_available()
        if not (torch.cuda.is_available() and torch.cuda.device_count() >= len(device.replace(',', ''))):
            LOGGER.info(s)
            install = 'See https://pytorch.org/get-started/locally/ for up-to-date torch install instructions if no ' \
                      'CUDA devices are seen by torch.\n' if torch.cuda.device_count() == 0 else ''
            raise ValueError(f"Invalid CUDA 'device={device}' requested."
                             f" Use 'device=cpu' or pass valid CUDA device(s) if available,"
                             f" i.e. 'device=0' or 'device=0,1,2,3' for Multi-GPU.\n"
                             f'\ntorch.cuda.is_available(): {torch.cuda.is_available()}'
                             f'\ntorch.cuda.device_count(): {torch.cuda.device_count()}'
                             f"\nos.environ['CUDA_VISIBLE_DEVICES']: {visible}\n"
                             f'{install}')

    if not cpu and not mps and torch.cuda.is_available():  # prefer GPU if available
        devices = device.split(',') if device else '0'  # range(torch.cuda.device_count())  # i.e. 0,1,6,7
        n = len(devices)  # device count
        if n > 1 and batch > 0 and batch % n != 0:  # check batch_size is divisible by device_count
            raise ValueError(f"'batch={batch}' must be a multiple of GPU count {n}. Try 'batch={batch // n * n}' or "
                             f"'batch={batch // n * n + n}', the nearest batch sizes evenly divisible by {n}.")
        space = ' ' * (len(s) + 1)
        for i, d in enumerate(devices):
            p = torch.cuda.get_device_properties(i)
            s += f"{'' if i == 0 else space}CUDA:{d} ({p.name}, {p.total_memory / (1 << 20):.0f}MiB)\n"  # bytes to MB
        arg = 'cuda:0'
    elif mps and getattr(torch, 'has_mps', False) and torch.backends.mps.is_available() and TORCH_2_0:
        # Prefer MPS if available
        s += 'MPS\n'
        arg = 'mps'
    else:  # revert to CPU
        s += 'CPU\n'
        arg = 'cpu'

    return torch.device(arg)

def intersect_dicts(da, db, exclude=()):
    """Returns a dictionary of intersecting keys with matching shapes, excluding 'exclude' keys, using da values."""
    return {k: v for k, v in da.items() if k in db and all(x not in k for x in exclude) and v.shape == db[k].shape}

def preprocess(img_pil, imgsz):
    img_pil = img_pil.resize((imgsz, imgsz))
    img_np = normalize_tensor(np.array(img_pil))
    img_np = np.transpose(img_np, [2, 0, 1])[np.newaxis, ...]
    return img_np

def normalize_tensor(tensor):
    return tensor/127.5 - 1
    # return tensor/255
    # return (tensor/255 - 0.44531356896770125 ) / 0.2692461874154524

def get_latest_opset():
    """Return second-most (for maturity) recently supported ONNX opset by this version of torch."""
    return max(int(k[14:]) for k in vars(torch.onnx) if 'symbolic_opset' in k) - 1  # opset


import os
from pathlib import Path
import shutil
import pandas as pd
import math

from PIL import Image, ImageDraw
import numpy as np
import matplotlib.pyplot as plt

from fdfat.utils.pose_estimation import MODEL_3D_POINTS

LMK_PARTS = [
    [0, 17], # jaw
    [17, 22], # left eye brown
    [22, 27], # right eye brown
    [27, 31], # nose
    [31, 36], # nose tip
    [36, 42], # left eye
    [42, 48], # right eye
    [48, 68], # mount
    [68, 70], # purpils
]

LMK_PART_NAMES = [
    "jaw", "leyeb", "reyeb", "nose", "nosetip", "leye", "reye", "mount", "purpils"
]

def circle(cx, cy, R=2):
    return cx - R//2, cy - R//2, cx + R//2, cy + R//2

def render_lmk_nme(nme, imgsz=1024, base_radius=200, title=None):

    # load model
    model_points = np.array(MODEL_3D_POINTS, dtype=np.float32)
    model_points = np.reshape(model_points, (3, -1)).T
    model_points[:, 2] *= -1

    # normalize
    minx, miny, maxx, maxy = model_points[:,0].min(), model_points[:,1].min(), model_points[:,0].max(), model_points[:,1].max()
    lmk_model = model_points[:,:2].copy()
    lmk_model[:,0] = (lmk_model[:,0] - minx)/(maxx-minx)
    lmk_model[:,1] = (lmk_model[:,1] - miny)/(maxy-miny)

    target = Image.new('RGB', (imgsz, imgsz), color=(255, 255, 255))
    draw = ImageDraw.Draw(target)

    lmk_image = lmk_model * (imgsz - 144) + 72
    for (x, y), v in zip(lmk_image, nme):
        target_radius = 200*v
        draw.ellipse(circle(x, y, R=target_radius), fill=(201, 168, 75))

        textbox = draw.textbbox((x, y+target_radius//2 + 2), f"{v*100:3.2f}")
        text_width = textbox[2]-textbox[0]
        draw.text((x-text_width//2, y+target_radius//2 + 2), f"{v*100:3.2f}", fill=(0, 0, 0))

    if title is not None:
        draw.text((10, 10), title, fill=(0, 0, 0))

    return target

def render_lmk(img, lmk, point_size=2, render=False):

    draw = ImageDraw.Draw(img)

    for begin, end in LMK_PARTS[:-1]:
        lx, ly = lmk[begin]
        for idx in range(begin+1, end):
            x, y = lmk[idx]
            draw.line([lx, ly, x, y], width=2, fill=(255, 0, 0))
            lx, ly = x, y

    for x, y in lmk:
        draw.rectangle([x-point_size/2, y-point_size/2, x+point_size/2, y+point_size/2], fill=(255, 255, 0))

    # bbox = gen_bbox(lmk)

    # print(bbox.flatten().astype(np.int32).tolist())
    # draw.rectangle(bbox.flatten().astype(np.int32).tolist(), width=2, outline=(0, 255, 255))

    if render:
        plt.imshow(img)
        plt.show()

    return img


def increment_path(path, exist_ok=False, sep='', mkdir=False):
    """
    Increments a file or directory path, i.e. runs/exp --> runs/exp{sep}2, runs/exp{sep}3, ... etc.

    If the path exists and exist_ok is not set to True, the path will be incremented by appending a number and sep to
    the end of the path. If the path is a file, the file extension will be preserved. If the path is a directory, the
    number will be appended directly to the end of the path. If mkdir is set to True, the path will be created as a
    directory if it does not already exist.

    Args:
        path (str, pathlib.Path): Path to increment.
        exist_ok (bool, optional): If True, the path will not be incremented and returned as-is. Defaults to False.
        sep (str, optional): Separator to use between the path and the incrementation number. Defaults to ''.
        mkdir (bool, optional): Create a directory if it does not exist. Defaults to False.

    Returns:
        (pathlib.Path): Incremented path.
    """
    path = Path(path)  # os-agnostic
    if path.exists() and not exist_ok:
        path, suffix = (path.with_suffix(''), path.suffix) if path.is_file() else (path, '')

        # Method 1
        for n in range(2, 9999):
            p = f'{path}{sep}{n}{suffix}'  # increment path
            if not os.path.exists(p):  #
                break
        path = Path(p)

    if mkdir:
        path.mkdir(parents=True, exist_ok=True)  # make directory

    return path

def make_dirs(dir='new_dir/'):
    # Create folders
    dir = Path(dir)
    if dir.exists():
        shutil.rmtree(dir)  # delete dir
    for p in dir, dir / 'labels', dir / 'images':
        p.mkdir(parents=True, exist_ok=True)  # make dir
    return dir

def generate_graph(csv_path, save_path, highlight_total=True):
    df = pd.read_csv(csv_path, sep="\t")
    all_fields = ["total", *LMK_PART_NAMES]
    epoch_idxes = df["epoch"].to_numpy()

    fig = plt.figure(figsize=(10,20))
    (ax1, ax2, ax3) = fig.subplots(3, 1, sharex=True, sharey=True)

    for f in all_fields:
        lw = 3 if highlight_total and f == "total" else 1
        ax1.plot(epoch_idxes, df[f].to_numpy(), label=f, linewidth=lw)
    ax1.legend()
    ax1.set_title("train")

    for f in all_fields:
        lw = 3 if highlight_total and f == "total" else 1
        ax2.plot(epoch_idxes, df[f"test_{f}"].to_numpy(), label=f, linewidth=lw)
    ax2.legend()
    ax2.set_title("val")
    
    ax3.plot(epoch_idxes, df[f"total"].to_numpy(), label="train", linewidth=3)
    ax3.plot(epoch_idxes, df[f"test_total"].to_numpy(), label="val", linewidth=3)
    ax3.legend()
    ax3.set_title("train + val")

    fig.savefig(save_path, bbox_inches='tight')
    plt.close()

def render_batch(batchx, batchy, save_path):
    batch_size = batchx.shape[0]

    grid_w = int(math.sqrt(batch_size)) + 1

    fig = plt.figure(figsize=(20,20))
    axes = fig.subplots(grid_w, grid_w, sharex=True, sharey=True, squeeze=False)
    for idx in range(batch_size):
        img_np = (batchx[idx,...].transpose([1, 2, 0]) * 127.5) + 127.5
        img = Image.fromarray(img_np.astype(np.uint8))
        lmk = (batchy[idx,:].reshape(70,2) + 0.5) * img.size[0]
        rendered = render_lmk(img, lmk)

        x = int(idx/grid_w)
        y = idx % grid_w
        axes[x, y].imshow(rendered)
        axes[x, y].axis('off')

    fig.savefig(save_path, bbox_inches='tight')
    plt.close()

def read_file_list(p, base_path=None):
    base_path = base_path if base_path is not None else ""
    with open(p, "r") as f:
        d = f.readlines()
        d = [os.path.join(base_path, a.strip("\n")) for a in d]
    return d



import time
import contextlib

class Profile(contextlib.ContextDecorator):
    # YOLOv5 Profile class. Usage: @Profile() decorator or 'with Profile():' context manager
    def __init__(self, name, max_sample=100):
        self.name = name
        self.samples = []
        self.max_sample=max_sample

    def __enter__(self):
        self.start = self.__time()
        return self

    def __exit__(self, type, value, traceback):
        t = self.__time() - self.start  # delta-time
        self.insert_time(t)

    def __time(self):
        return time.time()*1000

    def insert_time(self, value):
        self.dt = value
        self.samples.append(self.dt)
        if len(self.samples) > self.max_sample:
            self.samples.pop(0)

    def avg(self):
        if len(self.samples) == 0:
            return -1
        total = sum(self.samples)
        sizee = len(self.samples)
        target_t = int(total/sizee)
        return target_t

    def report(self):
        return self.avg()

    def print_report(self):
        print(f"[{int(time.time()*1000)}] {self.name}: {self.report()}ms")